------------------------------------------------------------------------------------------------
---  cron - a software package written in Lua for the Shards Online Server
--   Author: Yorlik
------------------------------------------------------------------------------------------------
-- TERMS OF USE: 
-- This file is distributed under the GNU Affero License Version 3  which can be seen here:
-- http://www.gnu.org/licenses/agpl-3.0.txt
-- You are not allowed to use this file if you do not have read and understood this license. 
------------------------------------------------------------------------------------------------

--- "Cron" is a behavor to implement a "cron" like service for other game objects ("Subscribers").
-- These subscribers can call the object this behavior is attached to by sending messages to it.
-- The object with the cron behavior on it is referred to as the "cron object", 
-- Usually you shouldn't need more than one cron objects per server.
-- All scheduling is done in game time,  not server- or real-time.
-- Subscribvers ask to get pinged at a certain time and do this  by 
-- SendMessage("CronSubscribe", {Id = number (usually "this.Id"), Interval=number ping_interval in game mionutes, StartTime={number year, number month, number day, number hour, number minute}, Name=string TimerName, Data = any arbitrary data)  
-- Example: ("cronObj" is the cron object)
-- cronObj:SendMessage("CronSubscribe", {Id=this.Id, Interval=15, StartTime={1023, 11, 29, 00, 5}, Name="timer01", Data = "Do Stuff!"}) 
-- Explanation: The message above would be sent to the object "cronObj" with a message name ("CronSubscribe"=mandatory) and  a table 
-- containing the data interpreted by the function handling the "CronSubscribe" message on the cron object.  
-- Id is the Id of the subscribing object, referred to as "this.Id" when used from the subscriber.
-- Interval is the interval in minutes gametime (not real time) at which the object wants to get pinged regularly, 
-- starting at StartTime defined by a table containing a list of numbers in the order {Year, Month, Day, Hour, Minute}.
-- Name is a string defining the timer, since an object can use several timers and so they can be distinguished.
-- Data is any arbitrary data which gets sent back to the object together with the ping. Can be number, string or even table or whatever.
---------------------------------------------------------------------------------------
--             Subscriber MUST implement the following Eventhandler:
---------------------------------------------------------------------------------------
-- To get pinged the subscribing object must define an Eventhandler in the following way:
-- RegisterEventHandler(EventType.Message,"cron" .. tostring(this.Id) .. "_".."MyTimerName", myHandlingFunction)
-- The format of the messagestring is mandatory: 
-- "cron" + Subscriber-Id + "_" + "Name of the timer as in the subscription message".
-- Or inline, like in this example: 
--[[
	RegisterEventHandler(EventType.Message,"cron" .. tostring(this.Id) .. "_".."timer01",
		function (Data)
      	-- example reaction to getting pinged:
		DebugMessage(this:GetName()..": I got pinged at Timer01! Data: "..tostring(Data), true)
		end
	)
]]--

----------------------------------------------------------------------------------
-- There are more messages (= usable cron functions) listed here by examples:
----------------------------------------------------------------------------------
-- For Subscription to a timer: 
-- 		cronObj:SendMessage("CronSubscribe",   {Id=this.Id, Interval=15, StartTime={1023, 11, 29, 00, 5}, Name="timer01", Data = "Do Stuff 01!"})
-- For UN-Subscribing a timer:
-- 		cronObj:SendMessage("CronUnSubscribe", {Id=this.Id, Name="timer01"}) 
-- For Un-Subscribing ALL timers the object has subscribed to:
-- 		cronObj:SendMessage("CronRemoveObject", this.Id) 
	    

---------------------------------------------------------------------------------- 
--[[
How the cron service works:
---------------------------------------------------------------------------------- 
The data sent by the CronSubscribe message contains a record which gets stored in a table called "crontab"
This table also gets stored as an Object Variable on the Cron Object and thus is persistent over server restarts.
On startup the cron object does some checks and also checks which of the entries are due to getting pinged.
Then a copy of the crontab entry is stored in another table called "SCHEDULE" along with a number containing the 
gametime in the format of game minutes since midnight. It is planned to extend this to a calender aware behavior.
A SCHEDULE entry looks like {200, {crontab_entry}}, where "crontab_entry" is a symbolic representation of the crontab entry record.
The example above would be pinged next at 3:20 gametime (=200 minutes since in game midnight)
The full record would look like: 
{195,  {Id=15485, Interval=15, StartTime={1023, 11, 29, 1, 5}, Name="timer01", Data = "Do Stuff!"}   }
After putting all entries into the SCHEDULE it gets sorted by the time to ping, the delay gets calculated and the timer fires.
Once the timer calls the eventhandler inside the cron object the first entry gets executed, removed and a new entry added 
with the new time to fire adjusted. So - the above record would look like this after being fired and adjusted:
{215,  {Id=15485, Interval=15, StartTime={1023, 11, 29, 1, 5}, Name="timer01", Data = "Do Stuff!"}   }
which would be 3:35 game time, 15 minutes later like before, sequence starting at the given starttime. 
]]-- 
----------------------------------------------------------------------------------

----------------------------------------
--- Basic crontab entry format: 
--- An entry is a table, so - 
--- the crontab is a table of tables
----------------------------------------
-- {Id=this.Id, Interval=2, StartTime={1023, 11, 29, 16, 59}, Name="timer01", Data = "Do Stuff 01!"}
-- this.Id = subscribing object handle
-- Interval = Interval at getting pinged in game minutes
-- StartTime = e.g. {1023, 11, 29, 16, 59} Server time in format {year, month, day, hour, minute} when to start the scheduling
--   The example would start November 29 in 1023, 16:59 game time
--	 in the moment only hour and minute are being used
-- Name = e.g. "myTimer01"  string with the name of the timer
-- Data = e.g. "Do Stuff 01!  - extra data which get sent back to the subscribing object and which can be used later. Can have any type, even table.

--------------------------
-- Startup
--------------------------
	require 'incl_gametime' 
	--- All the stuff to keep this file cleaner (incl_cronfuncs.lua)
	require 'incl_cronfuncs' -- include this second


--- Cron Globals
	TIMELEAP_TOLERANCE=5  -- Maximum allowed timeleap before dropping events

	-- Servers time compression factor 
	-- USAGE:
	-- Multiply Game Time with it to get Real Time
	-- Divide Real Time by it to get Game Time
	FACTOR=GetTotalDayDurationSecs()/(24*3600)

	-- Every 15 minutes 
 	-- Checks for removed objects and removes their entries
	HOUSEKEEPING_INTERVAL=15 * 60 * 1000


-- TODO --------------------------------------------------------

-- Interactive Test Functionality

----------------------------------------------------------
RegisterEventHandler(EventType.ClientObjectCommand, "interact", 
		function(user)
		user:SystemMessage("You attempt to interact: "..tostring(user))
		local menuItems = {
		"Show crontab",
		"Show first SCHEDULE entry",
		"Show Timers",
		"Un-Subscribe Timer 2",
	    "Un-Subscribe All"
		}
		user:RequestContextMenu(this,"cronobject_interact", menuItems)	
	end)


RegisterEventHandler(EventType.ContextMenuResponse,"cronobject_interact", 
	function (user,menuIndex)
	    if( menuIndex == 0 ) then return end
	    if( user == nil or not(user:IsValid())) then return end

	    --if( user:DistanceFrom(this) > OBJECT_INTERACTION_RANGE) then return end

	    if    ( menuIndex == 1 ) then
	    	
	    elseif( menuIndex == 2 ) then	
	    	
	    elseif( menuIndex == 3 ) then
	    	
	    elseif( menuIndex == 4 ) then
	    	
	    elseif( menuIndex == 5 ) then
	    	
	    end
	end)



--- Start Cron
CronStartup()










